import tempfile
from contextlib import redirect_stdout, redirect_stderr
from apicall.printutils import is_unicode_string, is_json, pprint

UNICODE_STRING = b'\xe6\x97\xa5\xe6\x9c\xac\xe8\xaa\x9e'
NON_UNICODE_DATA = b'\x93\xfa\x96{\x8c\xea'
NON_UNICODE_DATA_HEX= \
    b'0000000 93 fa 96 7b 8c ea\n' \
    b'0000006\n'


def FakeTerminal(mode: str):
    f = tempfile.TemporaryFile(mode)
    # isatty() method is always returns True.
    f.isatty = lambda: True
    return f


class TestIsUnicodeString:
    def test_zero_length(self):
        assert is_unicode_string(b'')

    def test_unicode_string(self):
        assert is_unicode_string(UNICODE_STRING)

    def test_non_unicode_data(self):
        assert not is_unicode_string(NON_UNICODE_DATA)


class TestIsJson:
    def test_valid_string(self):
        assert is_json(b'{}')
        assert is_json(b'{"name": "foo"}')

    def test_invalid_json(self):
        assert not is_json(b'')
        assert not is_json(b'{')
        assert not is_json(b"{'name': 'foo'}")


class TestWriteBinaryData:
    def test_stdout_is_not_terminal(self):
        out = tempfile.TemporaryFile('w+t')
        err = FakeTerminal('w+t')
        with out, err:
            with redirect_stdout(out), redirect_stderr(err):
                pprint(UNICODE_STRING)

            out.seek(0)
            assert out.buffer.read() == UNICODE_STRING
            err.seek(0)
            assert err.read() == ''

    def test_unicode_string(self):
        out = FakeTerminal('w+t')
        err = FakeTerminal('w+t')
        with out, err:
            with redirect_stdout(out), redirect_stderr(err):
                pprint(UNICODE_STRING)

            out.seek(0)
            assert out.buffer.read() == UNICODE_STRING
            err.seek(0)
            assert err.read() == ''

    def test_non_unicode_data(self):
        out = FakeTerminal('w+t')
        err = FakeTerminal('w+t')
        with out, err:
            with redirect_stdout(out), redirect_stderr(err):
                pprint(NON_UNICODE_DATA)

            out.seek(0)
            assert out.buffer.read() == NON_UNICODE_DATA_HEX
            err.seek(0)
            assert err.read().startswith('WARNING: ')



import typing
import json
from dataclasses import dataclass

import jsonrpcclient

from . import config
from . import restapi


class Request(typing.NamedTuple):
    """
    The Request class represents a JSON-RPC v2 request.

    This class is intended to encapsulate data for requests in a structured
    format. It includes a method and optional parameters, which can be used
    to describe the details of a request in various scenarios such as API
    calls or internal command handling.

    :ivar method: Method name for JSON-RPC request.
    :ivar params: Optional parameters for method.
    """
    method: str
    params: typing.Any = None

    def to_json(self):
        return jsonrpcclient.request(self.method, params=self.params)


class ErrorResponse(Exception):
    """
    Represents an error response in the context of an RPC or API request.

    This class is used to encapsulate the details of an error response, providing
    access to both the raw JSON data representing the error and the structured error
    object.

    :ivar json: The JSON returned by the JSON-RPC 2.0 server.
    :ivar response: The structured error response object.
    """
    def __init__(self, res_json: str, res: jsonrpcclient.Error):
        self.json = res_json
        self.response = res


@dataclass(frozen=True)
class Endpoint:
    urls: typing.Tuple[str, ...]
    headers: typing.Tuple[config.HttpHeader, ...]
    basic: typing.Optional[config.BasicAuth]

    def send(
            self,
            req: Request,
            verbose: int = 0,
            logging_cb: typing.Optional[typing.Callable[[str], None]] = None,
    )->str:
        req_rpc = req.to_json()
        res_raw = restapi.Request(
            method='POST',
            urls=self.urls,
            headers=self.headers,
            basic=self.basic,
            data=json.dumps(req_rpc).encode('utf8'),
        ).fetch(verbose=verbose, logging_cb=logging_cb)
        res_rpc = jsonrpcclient.responses.parse(res_raw.json)
        Endpoint._check_response(res_raw.body, res_rpc)
        return res_raw.body

    @staticmethod
    def _check_response(res_json: str, res_rpc: typing.Any):
        if isinstance(res_rpc, jsonrpcclient.Ok):
            return
        if isinstance(res_rpc, jsonrpcclient.Error):
            raise ErrorResponse(res_json, res_rpc)
        else:
            for r in res_rpc:
                Endpoint._check_response(res_json, r)



"""引数のパース処理のテスト

主要なコマンドに対して、想定されるいくつかのパターンの引数を入力し、
想定通りにパースできているかテストする。
"""
import pytest
import apicall.config as config
import apicall.arguments as arg
from . import command_tester as ct

saved_config = None


def fake_config_save(path: str, conf: config.Config):
    global saved_config
    saved_config = conf


@pytest.fixture(scope='function', autouse=True)
def patch_config():
    """ apicall.config.save()を置き換える

    save_config変数の初期化と、save()関数をfake_config_save()に置き換えをする。
    このパッチを当てると、テストケース実行時に、実際のファイルが作られてしまう問題を解消できる。

    この初期化処理は、テスト関数(メソッド)を呼び出す前に行われる。
    """
    global saved_config
    saved_config = None
    config.save = fake_config_save


class TestApicallCommand:
    def test_apicall_without_args(self):
        pc = ct.StartCondition(
            args=['apicall'],
            config=None,
        ).parse()
        assert pc.success
        assert pc.error_message == ''
        assert isinstance(pc.fn, arg.ApicallCommand)

        ec = pc.exec()
        assert ec.out.startswith('usage: apicall ')
        assert ec.err == ''
        assert ec.exit_code == arg.ExitInvalidArgs

    def test_restcall_without_args(self):
        pc = ct.StartCondition(
            args=['restcall'],
            config=None,
        ).parse()
        assert not pc.success
        assert pc.error_message.startswith('usage: restcall ')

    def test_jsonrpccall_without_args(self):
        pc = ct.StartCondition(
            args=['jsonrpccall'],
            config=None,
        ).parse()
        assert not pc.success
        assert pc.error_message.startswith('usage: jsonrpccall ')

    def test_auth_header_set(self):
        pc = ct.StartCondition(
            args=['apicall', 'auth', 'header', 'set', 'NAME', 'VALUE'],
            config=config.Config(),
        ).parse()
        assert pc.success
        assert pc.error_message == ''
        assert isinstance(pc.fn, arg.AuthHeaderSet)

        ec = pc.exec()
        assert ec.out == ''
        assert ec.err == ''
        assert ec.exit_code == arg.ExitOk
        assert saved_config == config.Config(
            headers=(config.HttpHeader('NAME', 'VALUE'), ), )

    def test_auth_header_unset(self):
        pc = ct.StartCondition(
            args=['apicall', 'auth', 'header', 'unset', 'HEADER2', 'HEADER3'],
            config=config.Config(
                headers=(
                    config.HttpHeader('header1', 'value1'),
                    config.HttpHeader('header2', 'value2'),
                    config.HttpHeader('header3', 'value3'),
                    config.HttpHeader('header4', 'value4'),
                ), ),
        ).parse()
        assert pc.success
        assert pc.error_message == ''
        assert isinstance(pc.fn, arg.AuthHeaderUnset)

        ec = pc.exec()
        assert ec.out == ''
        assert ec.err == ''
        assert ec.exit_code == arg.ExitOk
        assert saved_config == config.Config(
            headers=(
                config.HttpHeader('header1', 'value1'),
                config.HttpHeader('header4', 'value4'),
            ), )

    def test_auth_basic_set(self):
        pc = ct.StartCondition(
            args=['apicall', 'auth', 'basic', 'set', 'User', 'P@ssw0rd'],
            config=config.Config(),
        ).parse()
        assert pc.success
        assert pc.error_message == ''
        assert isinstance(pc.fn, arg.AuthBasicSet)

        ec = pc.exec()
        assert ec.out == ''
        assert ec.err == ''
        assert ec.exit_code == arg.ExitOk
        assert saved_config == config.Config(
            basic=config.BasicAuth(
                user='User',
                password='P@ssw0rd',
            ), )

    def test_auth_basic_unset(self):
        pc = ct.StartCondition(
            args=['apicall', 'auth', 'basic', 'unset'],
            config=config.Config(
                basic=config.BasicAuth(
                    user='User',
                    password='P@ssw0rd',
                ), )).parse()
        assert pc.success
        assert pc.error_message == ''
        assert isinstance(pc.fn, arg.AuthBasicUnset)

        ec = pc.exec()
        assert ec.out == ''
        assert ec.err == ''
        assert ec.exit_code == arg.ExitOk
        assert saved_config == config.Config()

    def test_endpoint_show(self):
        pc = ct.StartCondition(
            args=['apicall', 'endpoint'],
            config=config.Config(
                endpoints=('http://localhost:8080', ), )).parse()
        assert pc.success
        assert pc.error_message == ''
        assert isinstance(pc.fn, arg.Endpoint)

        ec = pc.exec()
        assert ec.out == 'http://localhost:8080\n'
        assert ec.err == ''
        # This operation MUST NOT update config file.
        assert saved_config == None

    def test_endpoint_set(self):
        pc = ct.StartCondition(
            args=['apicall', 'endpoint', 'http://localhost:3333'],
            config=config.Config(),
        ).parse()
        assert pc.success
        assert pc.error_message == ''
        assert isinstance(pc.fn, arg.Endpoint)

        ec = pc.exec()
        assert ec.out == ''
        assert ec.err == ''
        # This operation MUST NOT update config file.
        assert saved_config == config.Config(
            endpoints=('http://localhost:3333', ), )



from apicall.urlutils import is_absolute_url, escape_query, concat_urls


class TestIsAbsoluteUrl:
    def test_missing_schema(self):
        assert is_absolute_url('//localhost/path')

    def test_missing_hostname(self):
        assert not is_absolute_url('/path-only')

    def test_relative_path(self):
        assert not is_absolute_url('get')


class TestEscapeQuery:
    def test_normal(self):
        assert escape_query('name=value') == 'name=value'

    def test_escape_key(self):
        assert escape_query('a b=value') == 'a%20b=value'

    def test_escape_value(self):
        assert escape_query('name=a b') == 'name=a%20b'


class TestConcatUrls:
    def test_url_and_relative_path(self):
        assert concat_urls(
            endpoints=('http://example.com:8000/api/v1', ),
            url='users',
            queries=tuple(),
        ) == ('http://example.com:8000/api/v1/users', )

    def test_domain_and_path(self):
        assert concat_urls(
            endpoints=('http://example.com', ),
            url='api/v1/users',
            queries=tuple()) == ('http://example.com/api/v1/users', )

    def test_use_https(self):
        assert concat_urls(
            endpoints=('https://example.com', ),
            url='api/v1/users',
            queries=tuple()) == ('https://example.com/api/v1/users', )

    def test_use_abs_url(self):
        assert concat_urls(
            endpoints=('http://example.com:8000', ),
            url='https://localhost:3333',
            queries=tuple(),
        ) == ('https://localhost:3333', )
